from pyaes import AESModeOfOperationECB, Encrypter, Decrypter, PADDING_NONE
import hashlib
import base64
import logging

_LOGGER = logging.getLogger(__name__)


class AESCipher(object):
    """Helperclass to handle the en/decryption of the messages
    """
    def __init__(self, magic, access_key, password):
        self.bs = 16
        self.key = hashlib.md5(bytearray(access_key, "utf8") + magic).digest()
        self.key += hashlib.md5(magic + bytearray(password, "utf8")).digest()

    def encrypt(self, raw):
        if len(raw) % self.bs != 0:
            raw = self._pad(raw)
        cipher = Encrypter(AESModeOfOperationECB(self.key),
                           padding=PADDING_NONE)
        ciphertext = cipher.feed(raw) + cipher.feed()

        return base64.b64encode(ciphertext)

    def decrypt(self, enc):
        # trying to decrypt empty data fails
        if not enc:
            return ""
        enc = base64.b64decode(enc)
        cipher = Decrypter(AESModeOfOperationECB(self.key),
                           padding=PADDING_NONE)
        decrypted = cipher.feed(enc) + cipher.feed()
        try:
            r = decrypted.decode("utf8").rstrip(chr(0))
        except UnicodeDecodeError as e:
            raise
            
        return r

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(0)



import slixmpp
import asyncio
import logging
import ssl
from slixmpp.xmlstream import tostring

_LOGGER = logging.getLogger(__name__)

# Custom SSL Context to avoid blocking calls. Needs slixmpp >= 1.9.1
# Created with the same parameters as in slixmpp XMLStream class
SSL_CONTEXT = ssl.create_default_context()
SSL_CONTEXT.set_ciphers("DEFAULT")
SSL_CONTEXT.check_hostname = True
SSL_CONTEXT.verify_mode = ssl.CERT_REQUIRED
SSL_CONTEXT.set_default_verify_paths()

class NefitXmppClient(slixmpp.ClientXMPP):
    """XMPP client implementation using the slixmpp library
    """

    def __init__(self, jid, password, encryption, nefit_client=None):

        slixmpp.ClientXMPP.__init__(self, jid, password,
                                    sasl_mech="DIGEST-MD5",
                                    ssl_context=SSL_CONTEXT)
        self.nefit_client = nefit_client
        # the event signaling, that the connection is established
        self.connected_event = asyncio.Event()
        self.disconnected_event = asyncio.Event()
        self.message_event = asyncio.Event()

        # set the various callback handlers
        self.add_event_handler('session_start', self.session_start)
        self.add_event_handler('session_end', self.session_end)
        self.add_event_handler('failed_auth', self.on_failed_auth)
        self.add_event_handler('auth_success', self.on_auth_success)
        self.add_event_handler('message', self.message_callback)
        self.add_event_handler('carbon_received', self.carbonmsg_recv_callblack)
        self.add_event_handler('carbon_sent', self.carbonmsg_sent_callblack)
        self.register_plugin('xep_0199') #ping
        self.register_plugin('xep_0280') #carbons
        self.encryption = encryption

    def session_start(self, event):
        """Callback handler for the session start.
        """
        self.send_presence()
        self.get_roster()
        self['xep_0280'].enable()
        self.connected_event.set()

    async def session_end(self, event):
        _LOGGER.debug("Connection was closed")

        if self.nefit_client.session_end_callback:
            await self.nefit_client.session_end_callback()
            
        self.disconnected_event.set()

    async def on_failed_auth(self, event):
        """Callback handler for an unsuccessfull authentication.
        """
        if self.nefit_client.failed_auth_handler:
            await self.nefit_client.failed_auth_handler(event)
        else:
            _LOGGER.error('failed_auth event: %s', event)
            raise SystemError('Invalid login. Check credentials ' +
                              '(serial_number, access_key).')

    def on_auth_success(self, event):
        """Callback handler for a successfull authentication.
        """
        _LOGGER.debug('auth_success event: %s', event)

    async def message_callback(self, msg):
        """Callback handler for a received message.
        """
        await self.nefit_client.raw_message_callback(msg, 'message')
        self.message_event.set()

    async def carbonmsg_recv_callblack(self, msg):
        """Callback handler for a received carbon message.
        """
        await self.nefit_client.raw_message_callback(msg['carbon_received'], 'carbon')

    def carbonmsg_sent_callblack(self, msg):
        """Callback handler for a sent carbon message.
        """
        _LOGGER.debug('Carbon sent: %s', msg['carbon_sent'])

    def send_message(self, mto, mbody, msubject=None, mtype=None,
                     mhtml=None, mfrom=None, mnick=None):

        # wild hack that is necessary
        body = mbody.replace("\r", "&#13;\n")
        message = self.make_message(mto=mto, mfrom=mfrom, mbody=body)
        message['lang'] = None
        str_data = tostring(message.xml, xmlns=message.stream.default_ns,
                            stream=message.stream,
                            top_level=True)
        str_data = str_data.replace("&amp;#13;", "&#13;")
        message.stream.send_raw(str_data)



import json
import logging
from .provider.pyaes_impl import AESCipher
from .provider.slixmpp_impl import NefitXmppClient


_LOGGER = logging.getLogger(__name__)

PRODUCT_IDS = {'7746901877': 'Junkers Control CT 100'}


class NefitCore(object):
    _accesskey_prefix = 'Ct7ZR03b_'
    _rrc_contact_prefix = 'rrccontact_'
    _rrc_gateway_prefix = 'rrcgateway_'
    _magic = bytearray.fromhex("58f18d70f667c9c79ef7de435bf0f9b1" +
                               "553bbb6e61816212ab80e5b0d351fbb1")

    container = {}

    def __init__(self, serial_number, access_key, password,
                 host='wa2-mz36-qrmzh6.bosch.de',
                 message_callback=None):
        """Constructor for the NefitCore client.

        :param serial_number:
        :param access_key:
        :param password:
        :param host:
        :param sasl_mech:
        """
        serial_number = str(serial_number)
        self.serial_number = serial_number
        self.password = password
        self.access_key = access_key
        self.message_callback = message_callback

        self.encryption = AESCipher(self._magic, access_key, password)

        identifier = serial_number + "@" + host
        self.jid = self._from = self._rrc_contact_prefix + identifier
        self._to = self._rrc_gateway_prefix + identifier

        _LOGGER.debug('Initializing XMPP client')
        self.xmppclient = NefitXmppClient(jid=self.jid,
                                          password=self._accesskey_prefix +
                                          access_key,
                                          encryption=self.encryption,
                                          nefit_client=self)

    async def connect(self):
        self.xmppclient.connect()

    async def disconnect(self):
        self.xmppclient.disconnect()

    async def raw_message_callback(self, msg, source):
        decodeSuccess = False
        if msg['type'] in ('chat', 'normal'):
            headers = msg['body'].split("\n")[:-1]
            body = msg['body'].split("\n")[-1:][0]
            _LOGGER.debug('headers: %s', headers)
            _LOGGER.debug('body: %s', body)

            try:
                 # Try to decode the message. Will fail if password is incorrect
                response = self.encryption.decrypt(body)
                _LOGGER.debug('response: %s', response)

                statusline = headers[0]
                statuscode = statusline.split(" ")[1]
                if statuscode == "204":
                    _LOGGER.debug('Empty message (204 No Content) received')

                elif statuscode == "200":
                    try:
                        data = json.loads(response)
                        if self.message_callback:
                            await self.message_callback(data)
                    except json.decoder.JSONDecodeError as e:
                        _LOGGER.error('Error parsing message %s', msg)
                        _LOGGER.error(e)
                        return
                else:
                    _LOGGER.error("Errormessage received: %s", statusline)
                    if response:
                        _LOGGER.error(response)
                    else:
                        _LOGGER.error("response is null")
                    raise SystemError(statusline)
            except UnicodeDecodeError:
                if source == 'message':
                    _LOGGER.error('Error decrypting message. Password incorrect?')
                    if self.failed_auth_handler:
                        await self.failed_auth_handler('auth_error_password')

    def get(self, path):
        """Construct a "GET command"

        cmd -- the command to send
        """
        msg = 'GET %s HTTP/1.1\rUser-Agent: NefitEasy\r\r' % path
        _LOGGER.debug('Sending msg...%s', msg.replace("\r", "\\r"))

        # dirty hack that is necessary
        msg = msg.replace("\r", "&#13;\n")
        self.xmppclient.send_message(mto=self._to, mbody=msg)

    def put(self, path, data):
        encrypted_data = self.encryption.encrypt(data)
        body = 'PUT %s HTTP/1.1\r' % path + \
               'Content-Type: application/json\r' + \
               'Content-Length: %i\r' % len(encrypted_data) + \
               'User-Agent: NefitEasy\r\r' + \
               '%s' % encrypted_data.decode('utf-8')
        self.xmppclient.send_message(mto=self._to, mbody=body)

    def put_value(self, path, value):
        # For string values, there a space after the : is not allowed
        data = json.dumps({'value': value}, separators=(',', ':'))
        _LOGGER.debug("PUT %s : %s", path, data)
        self.put(path, data)

    def set_usermode(self, mode):
        self.put_value('/heatingCircuits/hc1/usermode', mode)

    def set_temperature(self, temperature):
        t = float(temperature)
        self.put_value('/heatingCircuits/hc1/temperatureRoomManual', t)
        self.put_value('/heatingCircuits/hc1/manualTempOverride/status', 'on')
        self.put_value('/heatingCircuits/hc1/manualTempOverride/temperature',
                       t)
